from typing import Optional

import httpx

from langchain_agentql.const import EXTRACT_DATA_ENDPOINT, REQUEST_ORIGIN
from langchain_agentql.messages import (
    QUERY_PROMPT_REQUIRED_ERROR_MESSAGE,
    QUERY_PROMPT_EXCLUSIVE_ERROR_MESSAGE,
    UNAUTHORIZED_ERROR_MESSAGE,
)


def handle_http_error(e: httpx.HTTPStatusError) -> None:
    response = e.response
    if response.status_code == httpx.codes.UNAUTHORIZED:
        raise ValueError(UNAUTHORIZED_ERROR_MESSAGE) from e
    
    msg = response.text
    try:
        error_json = response.json()
        msg = (
            error_json["error_info"]
            if "error_info" in error_json
            else str(error_json)
        )
    except (ValueError, TypeError):
        msg = f"HTTP {e}."
    raise ValueError(msg) from e


def load_data(
    url: str,
    api_key: str,
    metadata: dict,
    params: dict,
    timeout: int,
    query: Optional[str] = None,
    prompt: Optional[str] = None,
) -> dict:
    if not query and not prompt:
        raise ValueError(QUERY_PROMPT_REQUIRED_ERROR_MESSAGE)
    if query and prompt:
        raise ValueError(QUERY_PROMPT_EXCLUSIVE_ERROR_MESSAGE)

    payload = {"url": url, "query": query, "prompt": prompt, "params": params, "metadata": metadata}

    headers = {
        "X-API-Key": f"{api_key}",
        "Content-Type": "application/json",
        "X-TF-Request-Origin": REQUEST_ORIGIN,
    }

    try:
        response = httpx.post(
            EXTRACT_DATA_ENDPOINT,
            headers=headers,
            json=payload,
            timeout=timeout,
        )
        response.raise_for_status()

    except httpx.HTTPStatusError as e:
        handle_http_error(e)
    else:
        return response.json()


async def aload_data(
    url: str,
    api_key: str,
    metadata: dict,
    params: dict,
    timeout: int,
    query: Optional[str] = None,
    prompt: Optional[str] = None,
) -> dict:
    if not query and not prompt:
        raise ValueError(QUERY_PROMPT_REQUIRED_ERROR_MESSAGE)
    if query and prompt:
        raise ValueError(QUERY_PROMPT_EXCLUSIVE_ERROR_MESSAGE)

    payload = {"url": url, "query": query, "prompt": prompt, "params": params, "metadata": metadata}

    headers = {
        "X-API-Key": f"{api_key}",
        "Content-Type": "application/json",
        "X-TF-Request-Origin": REQUEST_ORIGIN,
    }

    async with httpx.AsyncClient() as client:
        try:
            response = await client.post(
                EXTRACT_DATA_ENDPOINT,
                headers=headers,
                json=payload,
                timeout=timeout,
            )
            response.raise_for_status()

        except httpx.HTTPStatusError as e:
            handle_http_error(e)
        else:
            return response.json()



""" AgentQL get web element from browser tool """

from typing import Optional, Type

from langchain_community.tools.playwright.base import BaseBrowserTool
from langchain_core.callbacks import (
    AsyncCallbackManagerForToolRun,
    CallbackManagerForToolRun,
)
from pydantic import BaseModel, Field

from langchain_agentql.const import (
    DEFAULT_EXTRACT_ELEMENTS_TIMEOUT_SECONDS,
    DEFAULT_INCLUDE_HIDDEN_ELEMENTS,
    DEFAULT_RESPONSE_MODE,
    DEFAULT_WAIT_FOR_NETWORK_IDLE,
    REQUEST_ORIGIN
)
from langchain_agentql.utils import (
    _aget_current_agentql_page,
    _get_current_agentql_page
)
from langchain_agentql.messages import MISSING_BROWSER_ERROR_MESSAGE
from langchain_agentql.llm_descriptions import (
    GET_WEB_ELEMENT_BROWSER_TOOL_DESCRIPTION,
    PROMPT_FIELD_ELEMENT_DESCRIPTION
)


class GetWebElementBrowserToolInput(BaseModel):
    """Input schema for AgentQL get web element from browser tool."""

    prompt: str = Field(..., description=PROMPT_FIELD_ELEMENT_DESCRIPTION)


class GetWebElementBrowserTool(BaseBrowserTool):
    """AgentQL get web element from browser tool.

    Setup:
        Install ``langchain-agentql`` and set environment variable ``AGENTQL_API_KEY``.

        .. code-block:: bash

            pip install -U langchain-agentql
            export AGENTQL_API_KEY="your-api-key"

    Instantiation:
        .. code-block:: python

            # sync tool usage
            tool = GetWebElementBrowserTool(
                sync_browser=sync_browser,
            )

            # async tool usage
            tool = GetWebElementBrowserTool(
                async_browser=async_browser,
            )

    Navigate to the target page with Playwright Navigate Tool:
        .. code-block:: python

            NavigateTool(sync_browser=sync_browser).invoke({"url": "https://www.agentql.com/blog"})

    Invocation with args:
        .. code-block:: python

            selector = tool.invoke({'prompt': 'The next page navigation button'})

        .. code-block:: python

            "[tf623_id='191']"
    
    Usage with Playwright Click Tool:
        .. code-block:: python

            ClickTool(async_browser=async_browser, visible_only=False).invoke({'selector': selector})
    """  # noqa: E501

    name: str = "get_web_element_from_browser"
    description: str = GET_WEB_ELEMENT_BROWSER_TOOL_DESCRIPTION
    args_schema: Type[BaseModel] = GetWebElementBrowserToolInput

    timeout: int = DEFAULT_EXTRACT_ELEMENTS_TIMEOUT_SECONDS
    """The number of seconds to wait for a request before timing out. Defaults to 300."""
    wait_for_network_idle: bool = DEFAULT_WAIT_FOR_NETWORK_IDLE
    """Whether to wait until the network reaches a full idle state before executing. Defaults to `True`."""
    include_hidden: bool = DEFAULT_INCLUDE_HIDDEN_ELEMENTS
    """Whether to take into account visually hidden elements on the page. Defaults to `False`."""
    mode: str = DEFAULT_RESPONSE_MODE
    """'standard' uses deep data analysis, while 'fast' trades some depth of analysis for speed and is adequate for most usecases.
    Learn more about the modes in this guide: https://docs.agentql.com/accuracy/standard-mode. Defaults to 'fast'."""

    def _run(
        self, 
        prompt: str, 
        _: Optional[CallbackManagerForToolRun] = None
    ) -> str:
        if not self.sync_browser:
            raise ValueError(MISSING_BROWSER_ERROR_MESSAGE)
        
        page = _get_current_agentql_page(self.sync_browser)
        element = page.get_by_prompt(
            prompt,
            self.timeout,
            self.wait_for_network_idle,
            self.include_hidden,
            self.mode,
            request_origin=REQUEST_ORIGIN
        )
        tf_id = element.get_attribute("tf623_id")
        return f"[tf623_id='{tf_id}']"

    async def _arun(
        self, 
        prompt: str, 
        _: Optional[AsyncCallbackManagerForToolRun] = None
    ) -> str:
        if not self.async_browser:
            raise ValueError(MISSING_BROWSER_ERROR_MESSAGE)
        
        page = await _aget_current_agentql_page(self.async_browser)
        element = await page.get_by_prompt(
            prompt,
            self.timeout,
            self.wait_for_network_idle,
            self.include_hidden,
            self.mode,
            request_origin=REQUEST_ORIGIN
        )
        tf_id = await element.get_attribute("tf623_id")
        return f"[tf623_id='{tf_id}']"



""" AgentQL toolkit """

from typing import List

from langchain_community.agent_toolkits.playwright.toolkit import (
    PlayWrightBrowserToolkit,
)
from langchain_core.tools import BaseTool

from langchain_agentql.tools import (
    ExtractWebDataBrowserTool,
    GetWebElementBrowserTool
)


class AgentQLBrowserToolkit(PlayWrightBrowserToolkit):
    """AgentQL toolkit.

    Setup:
        Install ``langchain-agentql`` and set environment variable ``AGENTQL_API_KEY``.

        .. code-block:: bash

            pip install -U langchain-agentql
            export AGENTQL_API_KEY="your-api-key"

    Key init args:
        sync_browser: Optional[SyncBrowser]
            A synchronous browser instance for executing tools synchronously
        async_browser: Optional[AsyncBrowser]
            An asynchronous browser instance for executing tools asynchronously

    Instantiate:
        .. code-block:: python

            from langchain-agentql import AgentQLBrowserToolkit

            toolkit = AgentQLBrowserToolkit(
                sync_browser=sync_browser
            )

            or

            toolkit = AgentQLBrowserToolkit.from_browser(
                sync_browser=sync_browser
            )

    Tools:
        .. code-block:: python

            agentql_tools = toolkit.get_tools()

        .. code-block:: none

            [ExtractWebDataBrowserTool(sync_browser=<Browser type=<BrowserType name=chromium ...>),
            GetWebElementBrowserTool(sync_browser=<Browser type=<BrowserType name=chromium ...>)]

    Use within an agent (Must use async browser):
        .. code-block:: python

            from langgraph.prebuilt import create_react_agent

            # Add Playwright Navigation Tool to agent's tools
            navigate_tool = NavigateTool(async_browser=async_browser)
            browser_agent_tools = toolkit.get_tools() + [navigate_tool]

            # Query the data
            agent_executor = create_react_agent(llm, browser_agent_tools)

            prompt = "Extract the data from https://www.agentql.com/blog using the following AgentQL query: { pokemons[] { name price }}"

            events = agent_executor.astream(
                {"messages": [("user", prompt)]},
                stream_mode="values",
            )
            async for event in events:
                event["messages"][-1].pretty_print()

        .. code-block:: none

            ================================ Human Message =================================
            Extract the data from https://www.agentql.com/blog using the following AgentQL query: { posts[] { title }}
            ================================== Ai Message ==================================
            Tool Calls:
            navigate_browser (call_XXXX)
            Call ID: call_XXXX
            Args:
                url: https://www.agentql.com/blog
            ================================= Tool Message =================================
            Name: navigate_browser

            Navigating to https://www.agentql.com/blog returned status code 200
            ================================== Ai Message ==================================
            Tool Calls:
            extract_web_data_from_browser (call_XXXX)
            Call ID: call_XXXX
            Args:
                query: { posts[] { title }}
            ================================= Tool Message =================================
            Name: extract_web_data_from_browser

            {"posts": [
                {"title": "Launch Week Recap—make the web AI-ready"}, 
                {"title": "Automated web application testing with AI and Playwright"},
                ...
            ]}
            ================================== Ai Message ==================================
            Here are the titles of the blog posts extracted from the AgentQL blog:

            1. Launch Week Recap—make the web AI-ready
            2. Automated web application testing with AI and Playwright
            ...
    """  # noqa: E501

    def get_tools(self) -> List[BaseTool]:
        """Get the tools in the toolkit."""
        return [
          ExtractWebDataBrowserTool.from_browser(sync_browser=self.sync_browser, async_browser=self.async_browser),
          GetWebElementBrowserTool.from_browser(sync_browser=self.sync_browser, async_browser=self.async_browser)
        ]



from typing import Optional, List, TYPE_CHECKING

import agentql
from langchain_community.tools.playwright.utils import (
    aget_current_page,
    get_current_page
)
from langchain_community.tools.playwright.utils import create_sync_playwright_browser as create_sync_playwright_browser_from_tools

try:
    from playwright.async_api import Browser as AsyncBrowser
    from playwright.async_api import Page as AsyncPage
    from playwright.sync_api import Browser as SyncBrowser
    from playwright.sync_api import Page as SyncPage
except ImportError as e:
    raise ImportError(
        "Unable to import playwright. Please make sure playwright module is properly installed."
    ) from e

def _get_current_agentql_page(browser: SyncBrowser) -> SyncPage:
    """
    Get the current page of the browser.
    Args:
        browser: The browser to get the current page from.
    Returns:
        Page: The current page.
    """
    return agentql.wrap(get_current_page(browser))


async def _aget_current_agentql_page(browser: AsyncBrowser) -> AsyncPage:
    """
    Get the current page of the browser.
    Args:
        browser: The browser to get the current page from.
    Returns:
        Page: The current page.
    """
    return await agentql.wrap_async(await aget_current_page(browser))


def create_sync_playwright_browser(
    headless: bool = True, args: Optional[List[str]] = None
) -> SyncBrowser:
    """
    Create a sync playwright browser.
    Args:
        headless: Whether to run the browser in headless mode. Defaults to True.
        args: arguments to pass to browser.chromium.launch
    Returns:
        Browser: The playwright browser.
    """
    return create_sync_playwright_browser_from_tools(headless=headless, args=args)

# Note: playwright's create_async_playwright_browser() function does not work properly
# https://github.com/langchain-ai/langchain/issues/15605

async def create_async_playwright_browser(
    headless: bool = True, args: Optional[List[str]] = None
) -> AsyncBrowser:
    """
    Create an async playwright browser.
    Args:
        headless: Whether to run the browser in headless mode. Defaults to True.
        args: arguments to pass to browser.chromium.launch
    Returns:
        Browser: The playwright browser.
    """
    from playwright.async_api import async_playwright

    playwright = await async_playwright().start()
    browser = await playwright.chromium.launch(headless=headless, args=args)
    return browser



"""AgentQL document loader."""

import os
from typing import Iterator, Optional

from langchain_core.document_loaders.base import BaseLoader
from langchain_core.documents import Document

from langchain_agentql.const import DEFAULT_API_TIMEOUT_SECONDS
from langchain_agentql.load_data import load_data

from langchain_agentql.const import (
    DEFAULT_IS_STEALTH_MODE_ENABLED,
    DEFAULT_WAIT_FOR_PAGE_LOAD_SECONDS,
    DEFAULT_IS_SCROLL_TO_BOTTOM_ENABLED,
    DEFAULT_IS_SCREENSHOT_ENABLED,
    DEFAULT_RESPONSE_MODE,
)
from langchain_agentql.messages import UNSET_API_KEY_ERROR_MESSAGE


class AgentQLLoader(BaseLoader):
    """AgentQL document loader.

    Setup:
        Install ``langchain-agentql`` and set environment variable ``AGENTQL_API_KEY``.
        You can get API key from our Dev Portal: https://dev.agentql.com/

        .. code-block:: bash

            pip install -U langchain-agentql
            export AGENTQL_API_KEY="your-api-key"

    Instantiate:
        .. code-block:: python

            from langchain_community.document_loaders import AgentQLLoader

            loader = AgentQLLoader(
                url = "https://www.agentql.com/blog",
                query = "{ posts[] { title url date author } }",
                is_scroll_to_bottom_enabled = True
                # To learn all the params and their specs, visit: https://docs.agentql.com/rest-api/api-reference
            )

    Lazy load:
        .. code-block:: python

        docs = loader.load()
        print(docs)

        .. code-block:: python

        [Document(
            metadata={
                'request_id': 'xxxxxx-xxxx-xxxx-xxxx-xxxx',
                'generated_query': None,
                'screenshot': None},
            page_content="{
                'posts': [
                    {
                        'title': 'Launch Week Recap—make the web AI-ready', 
                        'url': 'https://www.agentql.com/blog/2024-launch-week-recap', 
                        'date': 'Nov 18, 2024', 
                        'author': 'Rachel-Lee Nabors'
                    }
                    ...
                ]
            }"
        ]
        
    """  # noqa: E501

    def __init__(
        self,
        url: str,
        query: str,
        api_key: Optional[str] = None,
        timeout: int = DEFAULT_API_TIMEOUT_SECONDS,
        is_stealth_mode_enabled: bool = DEFAULT_IS_STEALTH_MODE_ENABLED,
        wait_for: int = DEFAULT_WAIT_FOR_PAGE_LOAD_SECONDS,
        is_scroll_to_bottom_enabled: bool = DEFAULT_IS_SCROLL_TO_BOTTOM_ENABLED,
        mode: str = DEFAULT_RESPONSE_MODE,
        is_screenshot_enabled: bool = DEFAULT_IS_SCREENSHOT_ENABLED,
    ):
        """
        Initialize with API key and params.

        Args:
            url (str): The URL of the web page you want to extract data from.
            query (str): The AgentQL query to execute. Learn more at https://docs.agentql.com/agentql-query
            api_key (Optional[str]): AgentQL API key. You can create one at https://dev.agentql.com.
            timeout (int): Seconds to wait for a request. Defaults to 900.
            is_stealth_mode_enabled (boolean): Enable experimental anti-bot evasion strategies. May not work for all websites at all times. Defaults to `False`.
            wait_for (int): Wait time in seconds for page load (max 10 seconds). Defaults to 0.
            is_scroll_to_bottom_enabled (boolean): Whether to scroll to bottom of the page before extracting data. Defaults to `False`.
            mode (str): 'standard' uses deep data analysis, while 'fast' trades some depth of analysis for speed. Learn more at https://docs.agentql.com/accuracy/standard-mode. Defaults to 'fast'.
            is_screenshot_enabled (boolean): Whether to take a screenshot before extracting data. Returned in 'metadata' as a Base64 string. Defaults to `False`.

            Visit https://docs.agentql.com/rest-api/api-reference for more details.
        """
        self.url = url
        self.query = query

        self._api_key = api_key or os.getenv("AGENTQL_API_KEY")
        if not self._api_key:
            raise ValueError(UNSET_API_KEY_ERROR_MESSAGE)
        
        self.timeout = timeout

        self.params = {
            "wait_for": wait_for,
            "is_scroll_to_bottom_enabled": is_scroll_to_bottom_enabled,
            "mode": mode,
            "is_screenshot_enabled": is_screenshot_enabled,
        }

        self.metadata = {
            "experimental_stealth_mode_enabled": is_stealth_mode_enabled,
        }

    def lazy_load(self) -> Iterator[Document]:
        data = load_data(
            url=self.url,
            query=self.query,
            api_key=self._api_key,
            metadata=self.metadata,
            params=self.params,
            timeout=self.timeout,
        )
        yield Document(
            page_content=str(data["data"]),
            metadata=data["metadata"],
        )
